import asyncio

from django.conf import settings
from django.contrib.postgres.search import SearchQuery
from django.db import transaction
from django.db.models import F, Q, Window
from django.db.models.functions import DenseRank
from django.dispatch import receiver
import orjson
from pgvector.django import CosineDistance

from torque import models as torque_models
from torque.signals import search_filter, search_index_rebuilt, update_cache_document

from semantic_search.llm import llm
from semantic_search.models import SemanticSearchCacheDocument
from semantic_search.utils import (
    batch_queryset,
    filter_queries,
    build_semantic_summaries,
)

from semantic_search.embedding_cache import embeddings_cache

BATCH_SIZE = 20
MAX_CONCURRENCY = 5


@receiver(update_cache_document)
def update_semantic_cache_document(sender, **kwargs):
    cache_document = kwargs["cache_document"]
    document_dict = kwargs["document_dict"]

    with transaction.atomic():
        SemanticSearchCacheDocument.objects.filter(
            search_cache_document=cache_document
        ).delete()

        semantic_summaries = build_semantic_summaries(document_dict)

        for semantic_summary in semantic_summaries:
            embeddings = embeddings_cache.get(semantic_summary)

            if not embeddings:
                embeddings = llm.get_embeddings(semantic_summary)
                embeddings_cache.add(semantic_summary, embeddings)

            semantic_search_cache_documents = [
                SemanticSearchCacheDocument(
                    search_cache_document=cache_document,
                    data=semantic_summary,
                    data_embedding=embedding,
                )
                for embedding in embeddings
            ]

        SemanticSearchCacheDocument.objects.bulk_create(semantic_search_cache_documents)


def rebuild_semantic_search_index_batch(scds, wiki_config):
    semantic_summaries_to_fetch = []
    semantic_sc_documents = []

    for scd in scds:
        cached_document_dict = torque_models.DocumentDictCache.objects.get(
            document=scd.document, wiki_config=wiki_config
        )

        document_dict = orjson.loads(cached_document_dict.dictionary)

        semantic_summaries = build_semantic_summaries(document_dict["fields"])

        for semantic_summary in semantic_summaries:
            embeddings = embeddings_cache.get(semantic_summary)

            if embeddings:
                semantic_sc_documents.append(
                    SemanticSearchCacheDocument(
                        search_cache_document=scd,
                        data_embedding=embeddings,
                        data=semantic_summary,
                    )
                )
            else:
                semantic_summaries_to_fetch.append((scd, semantic_summary))

    llm_embeddings = llm.get_embeddings(
        [summary for _scd, summary in semantic_summaries_to_fetch]
    )

    for (scd, semantic_summary), embedding in zip(
        semantic_summaries_to_fetch, llm_embeddings
    ):
        embeddings_cache.add(semantic_summary, embedding)
        semantic_sc_documents.append(
            SemanticSearchCacheDocument(
                search_cache_document=scd,
                data_embedding=embedding,
                data=semantic_summary,
            )
        )
    return semantic_sc_documents

    #SemanticSearchCacheDocument.objects.bulk_create(semantic_sc_documents)


@receiver(search_index_rebuilt)
def rebuild_semantic_search_index(sender, **kwargs):
    wiki_config = kwargs["wiki_config"]

    search_cache_documents = torque_models.SearchCacheDocument.objects.filter(
        wiki_config=wiki_config
    ).select_related("document")

    import threading

    batches = list(batch_queryset(search_cache_documents, BATCH_SIZE))
    batch_chunks = [(batches[i::MAX_CONCURRENCY], i) for i in range(MAX_CONCURRENCY)]
    results = [[]] * MAX_CONCURRENCY

    def rebuild_batch_chunk(batch_chunk_pair):
        for scds in batch_chunk_pair[0]:
            results[batch_chunk_pair[1]].append(rebuild_semantic_search_index_batch(scds, wiki_config))

    threads = [
        threading.Thread(target=rebuild_batch_chunk, args=(chunk,))
        for chunk in batch_chunks if chunk[0]  # Only create threads for non-empty chunks
    ]

    for thread in threads:
        thread.start()

    sscds_to_add = []
    for thread in threads:
        thread.join()

    for result in results:
        for sscds in result:
            # If there was an error for whatever reason, and therefore sscds didn't get set
            # just don't add it and move on
            if sscds:
                sscds_to_add.extend(sscds)

    SemanticSearchCacheDocument.objects.bulk_create(sscds_to_add, batch_size=BATCH_SIZE)

@receiver(search_filter)
def semantic_filter(sender, **kwargs):
    similarity = getattr(settings, "SEMANTIC_SEARCH_SIMILARITY", 0.7)

    cache_documents = kwargs["cache_documents"]
    qs = kwargs.get("qs")
    include_snippets = kwargs.get("include_snippets", False)
    qs_without_negations = filter_queries(qs, keep_negations=False)

    if qs_without_negations:
        embeddings = llm.get_embeddings(qs_without_negations)

        distances = {}
        semantic_qs = Q()
        for i, embedding in enumerate(embeddings):
            distance_col_name = f"distance_{i}"
            distances[distance_col_name] = CosineDistance(
                "semantic_documents__data_embedding", embedding
            )
            semantic_qs |= Q(**{f"{distance_col_name}__lte": similarity})

        # filters phrases and negations from the queries,
        # what we're calling exact or advanced search
        filter_qs = Q()
        for q in filter_queries(qs):
            if q != "":
                filter_qs &= Q(data_vector=SearchQuery(q, search_type="websearch"))

        results = (
            cache_documents.alias(**distances)
            .annotate(score=1 - F("distance_0"))
            .filter(semantic_qs)
            .filter(filter_qs)
        )

        if include_snippets:
            results = results.annotate(snippets=F("semantic_documents__data"))

        return results



import pytest
from semantic_search.utils import (
    dict_to_markdown,
    filter_queries,
    build_semantic_summaries,
)


@pytest.mark.parametrize(
    "query,expected_result",
    [
        # should remove all words
        ("water in india", ""),
        # should keep negated words
        ("-test", "-test"),
        # should remove all words except negated
        ("water in india -test", "-test"),
        # should remove words, keep negated phrase
        (
            'water in india -"test 123"',
            '-"test 123"',
        ),
        # should not remove words when OR-ed
        ('water OR "test"', 'water OR "test"'),
        # should keeps a phrase
        ('water AND "test"', '"test"'),
        # should remove AND, group where none of the terms are retained
        (
            "(water AND test)",
            "",
        ),
        # should keep negated term in AND, retain group
        ("(water AND -test)", "( -test)"),
        # should remove NOT (search doesn't yet support NOT), non-negated word
        ("NOT water", ""),
        # should keep phrase, remove NOT
        (
            'NOT "test"',
            '"test"',
        ),
        # all together now
        (
            'title:water water in india -"is a resource" -goats "test"',
            '-"is a resource" -goats "test"',
        ),
        # should handle single quotes at start of search query
        (
            "'water in india'",
            "",
        ),
    ],
)
def test_filter_queries(query, expected_result):
    results = filter_queries([query])
    assert results[0] == expected_result, (
        f"Expected {expected_result}, but got {results}"
    )


@pytest.mark.parametrize(
    "query,expected_result",
    [
        ("-test", ""),
        (
            'title:water water in india -"is a resource" -goats "test"',
            'title:water water in india "test"',
        ),
        (
            "'water in india'",
            "water in india'",
        ),
    ],
)
def test_filter_queries_without_negations(query, expected_result):
    results = filter_queries([query], keep_negations=False)
    assert results[0] == expected_result, (
        f"Expected {expected_result}, but got {results}"
    )


@pytest.mark.parametrize(
    "document_dict,expected_summary",
    [
        ({"title": "Test Document"}, "# title\n\nTest Document\n\n"),
        (
            {"summary": "This is a test document."},
            "# summary\n\nThis is a test document.\n\n",
        ),
        (
            {
                "details": {
                    "author": "John Doe",
                    "date": "2023-10-01",
                }
            },
            "# details\n\n## author\n\nJohn Doe\n\n## date\n\n2023-10-01\n\n\n",
        ),
        ({"tags": ["test", "document"]}, "# tags\n\n* test\n* document\n\n"),
    ],
)
def test_dict_to_markdown(document_dict, expected_summary):
    summaries = dict_to_markdown(document_dict)
    assert summaries == expected_summary, (
        f"Expected {expected_summary}, but got {summaries}"
    )


def test_build_semantic_summaries():
    document_dict = {
        "title": "Test Document",
        "summary": "This is a test document.",
        "details": {
            "author": "John Doe",
            "date": "2023-10-01",
        },
        "tags": ["test", "document"],
    }
    summaries = build_semantic_summaries(document_dict)
    expected_summaries = [
        "# title  \nTest Document",
        "# summary  \nThis is a test document.",
        "# details  \n## author  \nJohn Doe  \n## date  \n2023-10-01",
        "# tags  \n* test\n* document",
    ]
    assert summaries == expected_summaries



import asyncio
import httpx
import orjson
from django.conf import settings
from tenacity import retry, stop_after_attempt, wait_random_exponential

TIMEOUT = 60.0 * 8
RETRIES = 5


class LLM:
    def __init__(self):
        self.api_base = getattr(settings, "SEMANTIC_SEARCH_EMBEDDING_API_BASE_URL")
        self.api_key = getattr(settings, "SEMANTIC_SEARCH_EMBEDDING_API_KEY")
        self.model_name = getattr(settings, "SEMANTIC_SEARCH_EMBEDDING_MODEL")
        self.dimensions = getattr(settings, "SEMANTIC_SEARCH_NUM_DIMENSIONS", 768)

    @retry(stop=stop_after_attempt(RETRIES), wait=wait_random_exponential(multiplier=1, min=4, max=60))
    async def aget_embeddings(self, texts, *, prompt_name=None) -> list[list[float]]:
        data = {"input": texts, "model": self.model_name, "dimensions": self.dimensions}
        if prompt_name:
            data["type"] = prompt_name

        async with httpx.AsyncClient(
            base_url=self.api_base,
            headers={
                "Authorization": f"Bearer {self.api_key}",
                "Content-Type": "application/json",
            },
            timeout=TIMEOUT,
        ) as client:
            response = await client.post("embeddings", content=orjson.dumps(data))
            response.raise_for_status()

        payload = orjson.loads(response.text)
        results = payload.get("data", [])
        embeddings = [result.get("embedding") for result in results]

        return embeddings

    def get_embeddings(self, texts, *, prompt_name=None) -> list[list[float]]:
        return asyncio.run((self.aget_embeddings(texts, prompt_name=prompt_name)))


llm = LLM()



import luqum.parser
import ply.lex as lex
import ply.yacc as yacc
from langchain_text_splitters import (
    MarkdownHeaderTextSplitter,
    RecursiveCharacterTextSplitter,
)

from semantic_search.transformers import (
    FilterQueryTransformer,
    WithoutNegationsTransformer,
)


# patch Luqum's error handling to handle unexpected tokens more gracefully
luqum.parser.t_error = lambda t: t.lexer.skip(1)
lexer = lex.lex(module=luqum.parser)
parser = yacc.yacc(module=luqum.parser)

headers_to_split_on = [
    ("#", "Header 1"),
]

token_splitter = MarkdownHeaderTextSplitter(headers_to_split_on, strip_headers=False)

max_tokens = 1000
chunk_overlap = 30
encoder_name = "cl100k_base"

text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
    encoding_name=encoder_name,
    chunk_size=max_tokens,
    chunk_overlap=chunk_overlap,
)


def filter_queries(qs, keep_negations=True):
    """
    Filters a list of queries to keep only phrases, e.g. "term", and
    negations, and some operators,
    e.g. `title:water water in india -"is a resource" -goats "test"`
    becomes `-"is a resource" -goats "test"`.
    Or just removes negations.
    """

    Transformer = (
        FilterQueryTransformer if keep_negations else WithoutNegationsTransformer
    )

    return [str(Transformer().visit(parser.parse(q))).strip() for q in qs]


def dict_to_markdown(data, level=1):
    markdown = ""
    for key, value in data.items():
        if value:
            markdown += f"{'#' * level} {key}\n\n"
            if isinstance(value, dict):
                markdown += dict_to_markdown(value, level + 1)
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        markdown += dict_to_markdown(item, level + 1)
                    else:
                        markdown += f"* {item}\n"
            else:
                markdown += f"{value}\n"
            markdown += "\n"
    return markdown


def batch_queryset(qs, batch_size):
    batch = []
    for obj in qs:
        batch.append(obj)
        if len(batch) == batch_size:
            yield batch
            batch = []
    if batch:
        yield batch


def build_semantic_summaries(document_dict):
    text = dict_to_markdown(document_dict)
    summaries = [
        doc.page_content
        for doc in text_splitter.split_documents(token_splitter.split_text(text))
        if doc.page_content.strip()
    ]
    return summaries
